/*
 *                                              rjohnson@analogic.com
 *   Copyright(c) 2001-2003   Richard B. Johnson
 *
 *   This program may be used for any purpose. However, its
 *   functionality is not guaranteed. If you make changes or
 *   bug-fixes, please supply them to the author.
 *
 *   Released  07-AUG-2001
 *   This was written several years ago and has been used every time I
 *   booted my Linux PPP system.
 *
 *   Originally I read modem initialization strings from a file. But,
 *   alas, privileged users modified the file so eventually I couldn't
 *   log in remotely. So, the initialization strings are now in the
 *   header file. That fixed the problem.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <malloc.h>
#include <signal.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <termios.h>
#include "site.h"

#define BUF_LEN   0x100
#define NAM_LEN   0x80
#define FAIL -1
#define ArraySize(a) (sizeof((a)) / sizeof((a)[0]))
static int hangup(int fd, void *buffer);
static int wait_rdy(int fd);
static int read_tt(int fd, char *buffer, size_t space);
static void write_tt(int fd, const char *buffer);
static void my_exit(void);
static void errors(const char *call, int line);
static void set_sig(int sig, sig_t func, int flags);
static const char subdir[]="/var/lock/subsys/";
#define ERRORS(s) errors(s, __LINE__-1)
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 *   Memory allocation.
 */
struct mem {
    char buf[BUF_LEN];
    char nam[NAM_LEN];
    struct termios io_mod;
    struct termios io_orig;
    int fd;
    };

static struct mem *mem;

/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 *  Common routine to set up the signal handlers.
 */
static void set_sig(int sig, sig_t func, int flags)
{
    struct sigaction sa;
    if(sigaction(sig, NULL, &sa) < 0)
        ERRORS("sigaction");
    sa.sa_flags   = flags;
    sa.sa_handler = func;
    if(sigaction(sig, &sa, NULL) < 0)
        ERRORS("sigaction");
    return;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 *  This waits for a character or characters to be ready or until
 *  one second has occurred. It returns TRUE if characters are available
 *  and FALSE if it timed out.
 */
static int wait_rdy(int fd)
{
     fd_set mask;
     struct timeval tim;

     FD_ZERO(&mask);
     FD_SET(fd, &mask);
     tim.tv_sec =  1L;
     tim.tv_usec = 0L;
     return(select(fd + 1, &mask, NULL, NULL, &tim));
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 *  This reads a buffer of modem characters, possibly one byte at a time.
 *  This was the only way I could get a reliable input from the modem when
 *  it takes a long time to respond to some commands such as "AT&W".
 *  It returns a string in the buffer, and the string length as a
 *  return value.
 */
static int read_tt(int fd, char *buffer, size_t space)
{
    char *start;
    ssize_t len;

    start = buffer;
    *buffer = (char) 0x00;
    while((wait_rdy(fd)) && (space))
    {
        if((len = read(fd, buffer, space)) > 0) 
        {
            space  -= len;
            buffer += len;
        }
    }
    *buffer = (char) 0x00;
    return (start - buffer);
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 *  This writes a string to the modem.
 */
static void write_tt(int fd, const char *buffer)
{
    size_t len;
    int ret;
    len = strlen(buffer);
    while(len)
    {
        if((ret = write(fd, buffer, len)) <= 0)
            ERRORS("write");
        buffer += ret;
        len    -= ret;
    } 
    return;       
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 *   This logs an error and then exits.
 */
static void errors(const char *call, int line)
{
    fprintf(stderr, "Error from line %d, file %s, call %s (%s)\n",
                     line, __FILE__, call, strerror(errno));
    exit(EXIT_FAILURE);
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 *  This attempts to hang up the modem. If the modem is correctly
 *  configured, it should happen.
 */
static int hangup(int fd, void *buffer)
{
    struct termios *modem;
    tcflag_t c_cflag;
    modem = (struct termios *) buffer;
    if((tcgetattr(fd, modem)) == FAIL)         /* Get terminal params     */
        ERRORS("tcgetattr");
    c_cflag = modem->c_cflag;                   /* Save baud rate, etc.    */
    modem->c_cflag = CLOCAL;                    /* Clear baud-rate bits    */ 
    if((tcsetattr(fd, TCSANOW, modem)) == FAIL) /* Drop DTR to the modem   */
        ERRORS("tcsetattr");
    (void)sleep(1);                              /* Wait for it to happen   */
    modem->c_cflag = c_cflag;                    /* Restore modified pars   */
    (void)tcsetattr(fd, TCSANOW, modem);         /* Set modem back on line  */
    (void)sleep(1);                              /* Note: Modem takes time! */
    if(wait_rdy(fd))                             /* Eat any modem garbage   */
        if(tcflush(fd, TCIOFLUSH) == FAIL)
            ERRORS("tcflush");
    return 0;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 *   Common exit point for deleting lock-file, etc.
 */
static void my_exit()
{
    (void)close(mem->fd);
    if(*mem->nam) (void)unlink(mem->nam);
    free(mem);
    return;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 * Entry
 */
int main(int cnt, char *argv[])
{
    char *eol;
    int resp;
    size_t i;

    if(cnt < 2)
    {
        fprintf(stderr, "Usage:\n");
        fprintf(stderr, "%s /dev/modem [lock-file]\n", argv[0]);
        exit(EXIT_FAILURE);
    }
    if((mem = (struct mem *) malloc(sizeof(struct mem))) == NULL)
    {
        (void)fprintf(stderr, "Can't allocate memory!\n");
        exit(EXIT_FAILURE);
    }
    memset(mem, 0x00, sizeof(struct mem));
    if(argv[2] != NULL)       /* Lock file, we can fork */
    {
        if(fork()) { sleep(1); return 0; }
        strcpy(mem->nam, subdir);
        eol = strchr(mem->nam, '/');
        while((eol = strchr(++eol, '/')) != NULL)
        {
            *eol = (char) 0x00;
            (void)mkdir(mem->nam, 0644);
            *eol = (char) '/';
        }
        strcat(mem->nam, argv[2]);
        (void)close(open(mem->nam, O_WRONLY|O_CREAT|O_TRUNC, 0644));
    }
    (void)atexit(my_exit);
    fprintf(stdout, "Initializing modem, %s\n", argv[1]);
    (void)fflush(stdout);
    set_sig(SIGHUP,  SIG_IGN, SA_INTERRUPT|SA_RESTART);
    set_sig(SIGTTOU, SIG_IGN, SA_INTERRUPT|SA_RESTART);
    set_sig(SIGTTIN, SIG_IGN, SA_INTERRUPT|SA_RESTART);
    (void)chdir("/dev");
    if((mem->fd = open(argv[1], O_RDWR|O_NDELAY)) == FAIL)
    {
        fprintf(stderr, "Failed to open %s (%s)\n",
                         argv[1], strerror(errno));
        exit(EXIT_FAILURE);
    }
    if(fcntl(mem->fd, F_SETFL, fcntl(mem->fd, F_GETFL) & ~O_NDELAY) == FAIL)
        ERRORS("fcntl");
    (void)sleep(1);
    if(tcgetattr(mem->fd, &mem->io_orig) == FAIL)    /* Get terminal params */
        ERRORS("tcgetattr");
    /*
     * Set to dumb RAW mode with no echo and no character interpretation.
     */
    memset(&mem->io_mod, 0x00, sizeof(mem->io_mod));
    mem->io_mod.c_cflag     = SPEED|CS8|CREAD|CLOCAL|HUPCL;
    mem->io_mod.c_iflag     = IGNBRK|IGNPAR;
    mem->io_mod.c_cc[VMIN]  = (cc_t) 1;
    mem->io_mod.c_cc[VTIME] = (cc_t) 1;
    if(tcsetattr(mem->fd, TCSANOW, &mem->io_mod) == FAIL) 
        ERRORS("tcsetattr");
    (void)hangup(mem->fd, mem->buf); 
    mem->io_mod.c_cflag     = SPEED|CS8|CREAD|HUPCL;
    if(tcsetattr(mem->fd, TCSANOW, &mem->io_mod) == FAIL) 
        ERRORS("tcsetattr");
    for(i=resp=0; i< ArraySize(init); i++)
    {
        if(strstr(init[i], "SLEEP"))
            (void)sleep(1);
        else
            write_tt(mem->fd, init[i]);
        if(read_tt(mem->fd, mem->buf, BUF_LEN))     /* Read/display any echo */
        {
            resp++;
            if((eol = strchr(mem->buf, '\r')) != NULL)
            {
                *eol++ = (char) '\n';
                *eol   = (char) 0;
            }
            fprintf(stdout, mem->buf);
            (void)fflush(stdout);
        }
    }
    if(resp)
        fprintf(stdout, "Done!\n");
    else
        fprintf(stdout, "Modem never responded!\n");
    mem->io_orig.c_cflag &= ~CBAUD;              /* Mask old baud-rate    */
    mem->io_orig.c_cflag |= SPEED;               /* Leave new baud-rate   */
    if(tcsetattr(mem->fd, TCSANOW, &mem->io_orig) == FAIL)
        ERRORS("tcsetattr");
    exit(EXIT_SUCCESS);
    return 0;    /* Phony */
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
