/*
 *   Created   07-AUG-2001                      Richard B. Johnson
 *                                              rjohnson@analogic.com
 *   Copyright(c) 2001-2003   Richard B. Johnson
 *
 *   This program may be used for any purpose. However, its
 *   functionality is not guaranteed. If you make changes or
 *   bug-fixes, please supply them to the author.
 *
 *   Released  07-AUG-2001
 *   Version 1.0
 *
 *   This is a 'getty', designed to monitor a RS-232C terminal or
 *   a modem. Once it detects that someone is trying to log in,
 *   it presents the standard prompt and executes the standard log-in
 *   program.  It borrows the idea first used by Digital for PDP and
 *   VAX log-ins. It is not confused by "CONNECT NNN" messages from a
 *   modem. It also does not auto-baud because most everybody uses
 *   fixed baud-rates now.
 *
 *   This works by waiting until somebody hits a control-key 3 times
 *   in a row within 1 second. Then, one second later, it presents
 *   the log-in prompt. Any control key may be used. Normally you
 *   use [Enter].
 *
 *   It is not required that modem control bits be enabled. Programs
 *   that might be executed after a log-in, such as the PPP daemon,
 *   may enable hardware flow-control. Since this doesn't require
 *   flow-control to work, it is ideal for making initial connections
 *   to get a PPP link working.
 *
 *   Unlike some, this getty forks a child to do the log-in. The
 *   parent waits for the child to exit, i.e., the user logs off,
 *   then hangs up the modem before it exits. Init will restart and
 *   the cycle continues. The ability to hang up the modem is very
 *   useful. If you connect a modem, using the standard getty, you
 *   probably noted that the time between closing the user's files
 *   and init's starting the getty again, is not sufficient to
 *   hang up the modem.
 *
 *   Since this might be visible to hackers, serious attempts have
 *   been made to prevent exploits. The only way to get to a shell
 *   is if the normal login program, typically /bin/login, succeeds
 *   with a proper login.  All possible errors and buffer lengths are
 *   checked. Also memory is not allocated on the stack.
 *
 *   14-MAR-2003          Richard B. Johnson
 *   Version 1.1
 *   Added a third parameter, "m". If this exists on the command-line,
 *   conventional RTS/CTS is enabled, plus the getty waits for a
 *   carrier detect as is usual with modem-gettys.
 *
 *   29-MAR-2003          Richard B. Johnson
 *   Version 1.2
 *   Added a fourth parameter, a string. If it exists, it is the
 *   name of a lock-file which must disappear before rs-getty will
 *   use the modem. This is so some startup utility can initialize
 *   the modem before rs-getty gets control. 
 * 
 */


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/signal.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/utsname.h>

#define ERRORS(s) errors((s),__LINE__-1)

#define FAIL -1
#define TRUE 1
#define ROOT 0
#define ArraySize(a) (sizeof(a) / sizeof(a[0]))
#define MAX_BEL 2
#define MAX_WAIT 8           /* Seconds to wait for a login   */
#define TIMEOUT 30           /* Seconds to wait for the modem */
#define BUF_LEN 0x1000 
#define NAM_LEN 0x0100
#if !defined(CBAUDEX)
#define CBAUDEX CBAUD
#endif

#include "tty.h" 
static unsigned char  tt[] = TERMIOS;

static const struct {
    int baud;
    speed_t speed;
    } rates[] = {
    {     0,    B0      },
    {    50,    B50     },
    {    75,    B75     },
    {   110,    B110    },
    {   134,    B134    },
    {   150,    B150    },
    {   200,    B200    },
    {   300,    B300    },
    {   600,    B600    },
    {  1200,    B1200   },
    {  1800,    B1800   },
    {  2400,    B2400   },
    {  4800,    B4800   },
    {  9600,    B9600   },
    { 19200,    B19200  },
    { 38400,    B38400  },
    { 57600,    B57600  },
    {115200,    B115200 },
    {230400,    B230400 },
    {460800,    B460800 }};

/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
static void errors(const char *call, int line);
int main(int argc, char *argv[]);
static speed_t xlate_baud(int baud);
static void reaper(int unused);
static void sig_hangup(int unused);
static void hangup(int fd);
static void cleanup(void);
static void wait_for_cr(void);
static void set_sig(int sig, sig_t func, int flags);
static char  login[]=LOGIN;
static const char issue[]="/etc/issue";
static const char lock[]="/var/lock/subsys/";
static const char run[]="/var/run/rs-getty-%c.pid";
static const char dev[]="/dev";
static const char que[]="?";
static const char space[]=" ";
static const char prompt[]="login:";
static const char bel[]={ 0x07, 0x07 };
static const char tim[]="... Timeout reading terminal\n";
const char Version[]="Version 1.2";
struct DATA {
    char buf[BUF_LEN];
    char sys[NAM_LEN];
    char run[NAM_LEN];
    char lock[NAM_LEN];
    const char *cp;
    struct utsname *uts;
    struct termios st;
    struct termios rt;
    struct timeval tv;
    int is;
    int len;
    int modem;
    int fd;
    size_t i;
    pid_t pid;
    speed_t baud;
    fd_set rfds;
    };

static struct DATA *dat;

/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 *  Create and initialize a table for the signals to be initialized.
 *  Add any signals I forgot to this table.
 */
typedef struct {
    int signal;            /* Signal value   */
    sig_t handler;         /* Signal handler */
    int flags;             /* Signal flags   */
    } SIGS;

const SIGS sigs[] = {
    { SIGHUP,    SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGINT,    SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGQUIT,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGILL,    SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGTRAP,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGABRT,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGIOT,    SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGBUS,    SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGFPE,    SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGSEGV,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGUSR2,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGPIPE,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGALRM,   SIG_IGN,    SA_INTERRUPT|SA_RESTART },
    { SIGTERM,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGCHLD,   reaper,     SA_INTERRUPT|SA_RESTART },
    { SIGSTKFLT, SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGCONT,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGTSTP,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGTTIN,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGTTOU,   SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGURG,    SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGWINCH,  SIG_DFL,    SA_INTERRUPT|SA_RESTART },
    { SIGIO,     SIG_DFL,    SA_INTERRUPT|SA_RESTART }};

/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 *  What to do when we get signaled to hang up.
 */
static void sig_hangup(int unused)
{
    hangup(dat->fd);
    exit(EXIT_FAILURE);
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 *   This cleans up upon exit. It sets the terminal back to sanity
 *   and deletes the pid file.
 */
static void cleanup()
{
    (void)tcsetattr(dat->fd, TCSANOW, (struct termios *) tt);
    if(*dat->run)  (void)unlink(dat->run);
    if(*dat->lock) (void)unlink(dat->lock);
    free(dat);
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 *  This waits for the user to enter MAX_BEL carriage returns and then
 *  returns.
 */ 
static void wait_for_cr()
{
    size_t i;
    for(i=0;;)
    {
        FD_ZERO(&dat->rfds);
        FD_SET(dat->fd, &dat->rfds);
        dat->tv.tv_usec = 0;
        dat->tv.tv_sec  = 1;
        if(!select(dat->fd+1, &dat->rfds, NULL, NULL, &dat->tv) && (i>MAX_BEL))
             break;
        if(FD_ISSET(dat->fd, &dat->rfds))
        {
            (void)read(dat->fd, dat->buf, BUF_LEN);
            if(*dat->buf < (char)' ')
            {
                if(++i > MAX_BEL )
                    (void)write(dat->fd, bel, sizeof(bel));
                continue;
            }
        }
        i = 0;
    }
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This gets the return status of any child processes and throws it
 *   away. In this manner, we don't retain any zombies.
 */
static void reaper(int unused)
{
    while(wait3(&unused, WNOHANG, NULL) > 0)
          ;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  This hangs up the modem. This usually produces some errors on
 *  Linux, depending upon the version. These errors are ignored.
 */
static void hangup(int fd)
{
    struct termios st;
    tcflag_t c_cflag;
    (void)tcgetattr(fd, &st);
    c_cflag = st.c_cflag;
    st.c_cflag = CLOCAL;
    (void)tcsetattr(fd, TCSADRAIN, &st);
    (void)sleep(1);
    st.c_cflag = c_cflag;
    (void)tcsetattr(fd, TCSANOW, &st);
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 *  This times out and hangs up the modem if the caller takes too much
 *  time typing in a name.
 */
static void catcher(int unused)
{
    (void)write(dat->fd, tim, sizeof(tim)-1);
    if(tcdrain(dat->fd) < 0)
        ERRORS("tcdrain");
    hangup(dat->fd);
    exit(EXIT_FAILURE);
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 *  Common routine to set up the signal handlers.
 */
static void set_sig(int sig, sig_t func, int flags)
{
    struct sigaction sa;
    if(sigaction(sig, NULL, &sa) < 0)
        ERRORS("sigaction");
    sa.sa_flags   = flags;
    sa.sa_handler = func;
    if(sigaction(sig, &sa, NULL) < 0)
        ERRORS("sigaction");
    return;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 *  This writes an error message to standard error then exits with
 *  the standard error code.
 */
static void errors(const char *call, int line)
{
    fprintf(stderr, "Error from line %d, file %s, call %s, %s\n", 
                    line, __FILE__, call, strerror(errno));
    exit(EXIT_FAILURE);
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 *  Translate the input binary speed to an output type of speed_t.
 */
static speed_t xlate_baud(int baud)
{
    size_t i;
    for(i = 0;  i < ArraySize(rates); i++)
        if(baud == rates[i].baud)
            return rates[i].speed;
    errno = EINVAL;
    return FAIL;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 *  There is no data allocated on the stack (no local data). This
 *  is the first line of defense against attempts to break into the
 *  system.
 */
int main(int args, char *argv[])
{
    errno = ENOMEM;                        /* Possible malloc problem */
    if((dat = (struct DATA *) malloc(sizeof(struct DATA))) == NULL)
        ERRORS("malloc");
    memset(dat, 0x00, sizeof(struct DATA));
    (void)atexit(cleanup);                 /* Cleanup upon exit       */
    dat->cp = que;
    if(argv[1] != NULL)   /* The tty */
    {
        dat->cp = argv[1];
        while( (*dat->cp)  && ((*dat->cp < '0') || (*dat->cp > '9')) )
             dat->cp++;
        if(!*dat->cp)
            dat->cp = que; 
    }
    sprintf(dat->run, run, *dat->cp);
    if((dat->fd = open(dat->run, O_WRONLY|O_CREAT|O_TRUNC, 0644)) == FAIL)
        ERRORS("open");
    (void)write(dat->fd, dat->buf, sprintf(dat->buf, "%d\n", getpid()));
/*
 *  Close ALL file descriptors! This will become a login process.
 */
    if((dat->fd = open("/", O_RDONLY)) == FAIL)
        ERRORS("open");
    while(dat->fd >= 0)
        (void)close(dat->fd--);
    if((dat->fd = open("/dev/console", O_RDWR)) != FAIL)
    {
        (void)dup2(dat->fd, STDIN_FILENO);
        (void)dup2(dat->fd, STDOUT_FILENO);
        (void)dup2(dat->fd, STDERR_FILENO);
    } 
    if(args < 3)
    {
        fprintf(stderr, "Usage:\n%s term baud [-m] [lockfile]\n", argv[0]);
        exit(EXIT_FAILURE);
    }
    if(argv[4] != NULL)
    {
        strcpy(dat->lock, lock);
        strcat(dat->lock, argv[4]);
        for(dat->i = 0; dat->i < TIMEOUT; dat->i++)
        {
             if(access(dat->lock, F_OK))
                 break;
             (void)sleep(1);
        }
        if((dat->fd = open(dat->lock, O_WRONLY|O_CREAT|O_TRUNC, 0644)) != FAIL)
            (void)close(dat->fd);
    }
    dat->modem = 0;
    if(argv[3] != NULL)
        if(strchr(argv[3], 'm') != NULL)
            dat->modem = TRUE;
    memcpy(&dat->st, tt,   sizeof(tt));
    set_sig(SIGCHLD, reaper,      SA_INTERRUPT|SA_RESTART);
    set_sig(SIGINT,  sig_hangup,  SA_INTERRUPT|SA_RESTART);
    set_sig(SIGHUP,  sig_hangup,  SA_INTERRUPT|SA_RESTART);
    dat->uts = (struct utsname *) dat->buf;
    if(uname(dat->uts) == FAIL)
        ERRORS("uname");
    strcpy(dat->sys, dat->uts->sysname);
    strcat(dat->sys, space);
    strcat(dat->sys, dat->uts->release);
    strcat(dat->sys, space);
    strcat(dat->sys, dat->uts->machine);
    strcat(dat->sys, space);
    strcat(dat->sys, dat->uts->nodename);
    strcat(dat->sys, space);
    strcat(dat->sys, prompt);
    strcat(dat->sys, space);
    (void)chdir(dev);
    if((dat->fd = open(argv[1], O_RDWR|O_NDELAY)) == FAIL)
        ERRORS("open");
    if(fcntl(dat->fd, F_SETFL, fcntl(dat->fd, F_GETFL) & ~O_NDELAY) == FAIL)
        ERRORS("fcntl");
    if((dat->baud = xlate_baud(atoi(argv[2]))) == FAIL)
        ERRORS("xlate_baud");
    if(chown(argv[1], ROOT, TTY) == FAIL)
        ERRORS("chown");
    if(chmod(argv[1], S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH) == FAIL)
        ERRORS("chmod");
    dat->st.c_cflag     = CS8|HUPCL|CREAD|CLOCAL|dat->baud;
    dat->st.c_cc[VMIN]  = 1;
    dat->st.c_cc[VTIME] = 0;
    dat->rt.c_cflag     = CS8|HUPCL|CREAD|CLOCAL|dat->baud;
    dat->rt.c_iflag     = IGNBRK|IGNPAR;
    dat->rt.c_oflag     = OPOST|ONLCR;
    dat->rt.c_cc[VMIN]  = 1;
    dat->rt.c_cc[VTIME] = 0;
    if(dat->modem)
    {
        dat->st.c_cflag |= CRTSCTS;
        dat->rt.c_cflag |= CRTSCTS;
        dat->st.c_cflag &= ~CLOCAL;
        dat->rt.c_cflag &= ~CLOCAL;
        if(tcsetattr(dat->fd, TCSANOW, &dat->rt) == FAIL)
            ERRORS("tcsetattr");
        if(tcflush(dat->fd, TCIOFLUSH) == FAIL)
            ERRORS("tcflush");
        (void)close(dat->fd);
        if((dat->fd = open(argv[1], O_RDWR)) == FAIL)
            ERRORS("open");
    }
    else
    {
        if(tcsetattr(dat->fd, TCSANOW, &dat->rt) == FAIL)
            ERRORS("tcsetattr");
        wait_for_cr();
    }
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 *  We are connected, the user entered the CR/CR/CR sequence or the
 *  modem carrier was detected. We now set the terminal for normal
 *  processing and write the contents of /etc/issue if it exists.
 */
    set_sig(SIGHUP, SIG_IGN, SA_INTERRUPT|SA_RESTART);
    if(tcsetattr(dat->fd, TCSANOW, &dat->st) == FAIL)
        ERRORS("tcsetattr");
    if(tcflush(dat->fd, TCIOFLUSH) == FAIL)
        ERRORS("tcflush");
    if((dat->is = open(issue, O_RDONLY)) != FAIL)
    {
        while((dat->len = read(dat->is, dat->buf, BUF_LEN)) > 0)
            if(write(dat->fd, dat->buf, (size_t) dat->len) != dat->len)
                ERRORS("write");
        if(close(dat->is) == FAIL)
            ERRORS("close");
    }
    dat->len = strlen(dat->sys);
    if(write(dat->fd, dat->sys, (size_t) dat->len) != dat->len)
        ERRORS("write");
    if(tcdrain(dat->fd) == FAIL)
        ERRORS("tcdrain"); 
    set_sig(SIGALRM, catcher, SA_INTERRUPT);
    (void)alarm(MAX_WAIT);
    if((dat->len = read(dat->fd, dat->buf, BUF_LEN)) <= 0)
        ERRORS("read");
    dat->buf[--dat->len] = (char) 0x00;
    (void)alarm(0);
    set_sig(SIGALRM, SIG_DFL, SA_INTERRUPT|SA_RESTART);
    switch(fork())
    {
    case 0:
        dat->pid = getpid();
        if(dup2(dat->fd, STDIN_FILENO)  != STDIN_FILENO)
            ERRORS("dup2");
        if(dup2(dat->fd, STDOUT_FILENO) != STDOUT_FILENO)
            ERRORS("dup2");
        if(dup2(dat->fd, STDERR_FILENO) != STDERR_FILENO)
            ERRORS("dup2");
        if(close(dat->fd) == FAIL)
            ERRORS("close");
        for(dat->i=0; dat->i < ArraySize(sigs); dat->i++)
            set_sig(sigs[dat->i].signal,
                    sigs[dat->i].handler,
                    sigs[dat->i].flags);
        if(setsid() == FAIL)
            ERRORS("setsid");
        if(fcntl(STDIN_FILENO, F_SETOWN, dat->pid) == FAIL)
            ERRORS("fcntl(F_SETOWN)");
        if(ioctl(STDIN_FILENO, TIOCSCTTY, TRUE) == FAIL)
            ERRORS("ioctl(TIOCSCTTY)");
        if(tcsetpgrp(STDIN_FILENO, dat->pid) == FAIL)
            ERRORS("tcsetpgrp");
        (void)execl(login, login, dat->buf, NULL);
        ERRORS("execl");
    case FAIL:
        ERRORS("fork");
    default:
        if(wait(&dat->is) == FAIL)
            ERRORS("wait");
    }
/*
 *  Hang up the modem (if it exists) before exit and restart. Because
 *  of some BUG that nobody admits to, I have to re-open the device.
 *  If I try to use the existing open fd to the modem, all operations
 *  return I/O errors.
 */
    if((dat->is = open(argv[1], O_RDWR|O_NDELAY)) == FAIL)
        ERRORS("open");
    if(tcsetattr(dat->is, TCSANOW, &dat->rt) == FAIL)
        ERRORS("tcsetattr");
    hangup(dat->is);
    if(close(dat->is) == FAIL)
        ERRORS("close");
    if(close(dat->fd) == FAIL)
        ERRORS("close");
    return 0;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=- E N D -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
